/*
    ftwall - Fast Track Firewall. This program allows iptables firewalls
             to filter traffic from Fast Track software such as Kazaa.

    Copyright (C) 2003, Chris Lowth

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    This product includes software developed by the OpenSSL Project
    for use in the OpenSSL Toolkit (http://www.openssl.org/)

    This product includes software written by Eric Young
    (eay@cryptsoft.com).  This product includes software written by Tim
    Hudson (tjh@cryptsoft.com).
*/

#include <libipq.h>
#ifndef _LIBIPQ_H
#error "+--------------------------------------------------------------------+"
#error "|                                                                    |"
#error "| NOTE: hibipq.h not found - Please refer to step 2 in the INSTALL   |"
#error "|       file - you MUST download and compile the iptables sources in |"
#error "|       the RIGHT location.                                          |"
#error "|                                                                    |"
#error "+--------------------------------------------------------------------+"
#else
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <assert.h>
#include <signal.h>
#include <syslog.h>
#include <limits.h>
#include "lhash/lhash.h"
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>

/* How much of the packet data payload to put on the QUEUE */
#define PAYLOADSIZE 512

/* How big the netlink buffer should be */
#define BUFSIZE (PAYLOADSIZE + 1024)

#define TRUE 1
#define FALSE 0

/* Some macros to make for shorter and more readable code */

#define SRC_ADDR (*(in_addr_t *)(payload+12))
#define DST_ADDR (*(in_addr_t *)(payload+16))

#define SRC_TCP_PORT (payload[ip_hlen] << 8 | payload[ip_hlen+1])
#define DST_TCP_PORT (payload[ip_hlen+2] << 8 | payload[ip_hlen+3])

#define SRC_UDP_PORT (payload[ip_hlen] << 8 | payload[ip_hlen+1])
#define DST_UDP_PORT (payload[ip_hlen+2] << 8 | payload[ip_hlen+3])

typedef unsigned char uchar;

typedef struct {		/* Structure describing something that is      */
	in_addr_t ip;		/* blocked. We keep IP address, port and timer */
	int port;		/* here. These are used differently by the     */
	time_t time;		/* different hashes.                           */
	int probed;
} BLOCKAGE;

LHASH *blockages_hash = NULL;	/* List of remote peers being blocked */

LHASH *clients_hash = NULL;	/* List of identified clients */

int probe_socket;		/* Socket used for sending UDP probes */

int raw_socket;			/* Socket used for sending fake replies */

FILE *FTWALL_STORE = NULL;		/* Peer address database file handle */

/* Configuration vars */

char *storefilename	= NULL;		/* change with "-f" */

char *active_client_directory = NULL;	/* change with "-D" */

int debug		= FALSE;	/* "-d" turns this on */

int stay_in_foreground  = FALSE;	/* "-s" or "-d" turns this on */

int block_udp		= TRUE;		/* "-n" turns this off */

int sock_bsize		= 32 * 1024;	/* change with "-b" */

time_t scan_timer	= 5;		/* How ofter to scan for idel clients */

time_t timelock_secs	= 120;		/* Disallow any TCP/IP syns for this
					   time after a KaZaA UDP.
					   change with "-t" */

int probe_port		= 1219;		/* Port we get probe replies on */

/* What information to write to the log file - change with "-l" or "-L"*/
#define LOG_DROPPED_UDP		0x0001
#define LOG_ACCEPTED_UDP	0x0002
#define LOG_DROPPED_TCP		0x0004
#define LOG_ACCEPTED_TCP	0x0008
#define LOG_IDENTIFIED_PEERS	0x0010
#define LOG_IDENTIFIED_CLIENTS	0x0020
int what2log		= 0;

/* Which log file to write to - change with "-l" or "-L" */
#define LOG_TO_SYSLOG		0x0001
#define LOG_TO_STDERR		0x0002
int log2		= 0;

char *hint = NULL;		/* If this is non-null, the text it points to
				   is added to error messages as hint about
				   how to fix the problem */

time_t now = 0;			/* The time(NULL) value when the current
				   packet arrived from netlink */

int stdio_is_null = FALSE;	/* Are the stdout/stdin/stderr handles connected
				   to /dev/null ? */

/* This is the openning UDP packet commonly sent by clients.
 * We send it too, as a kind of "ping" */
uchar probe_pkt[]=   {0x27,0x00,0x00,0x00,0x29,0x80,'K','a','Z','a',
	'A',0x00};

/* This is a "reply" packet that we send back when a client
 * tries to open a connection - to keep it from switching to
 * encrypted TCP/IP mode for as long as possible */
uchar reply_packet[]={0x28,0x00,0x00,0x00,0x29,0x00,0x39,0x9d,0x2a,
	0x01,0xef,'K','a','Z','a','A',0x00};
	
/*************************************************************/

static void error_msg(char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);

	if (!stay_in_foreground)
		vsyslog(LOG_ERR, fmt, ap);

	if (!stdio_is_null) {
		vfprintf(stderr, fmt, ap);
		fprintf(stderr, "\n");
	}
}

/****************************************************************************
 * die - curl up our toes and pass on into the next world, making some
 * minor effort to tidy up as we go.
 */

static void die(struct ipq_handle *h, char *s, ...)
{
	char msg[1024];
	char *e;
	va_list ap;

	va_start(ap, s);

	if (s)
		vsprintf(msg,s,ap);
	else
		strcpy(msg,"ERROR");

	/*
	 * I *really* dont like this test, but libipq makes its
	 * internal ipq_errno variable "static", so there's no
	 * way round it - beyond a patch (which I also dont like).
	 */

	e = ipq_errstr();
	if (memcmp(e, "Unknown", 7) != 0)
		sprintf(msg+strlen(msg), ": %s", e);

	if (errno)
		sprintf(msg+strlen(msg), ": %s", strerror(errno));

	if (hint != NULL)
		sprintf(msg+strlen(msg)," (%s)", hint);

	error_msg( "%s", msg );

	ipq_destroy_handle(h);

	exit(2);
}

/****************************************************************************
 * write_log( .. ) - Write a message to syslog or stderr, depending on what
 * has been configured via the command line.
 */

static void write_log( char *fmt, ... )
{
	va_list ap;

	va_start(ap, fmt);

	if (log2 & LOG_TO_STDERR) {
		vfprintf(stderr, fmt, ap);
		fprintf(stderr, "\n");
	}

	if (log2 & LOG_TO_SYSLOG)
		vsyslog(LOG_INFO, fmt, ap);
}

/****************************************************************************
 * print_addr( .. ) - output a readable, textual representation of an IP
 * address and port number. Typically only used for generating debug
 * output.
 */

static print_addr( FILE *f, in_addr_t ip, int port )
{
	struct in_addr a;
	a.s_addr = ip;
	fprintf(f, "%s:%d", inet_ntoa(a), port);
}

static void print_syn(FILE *f, in_addr_t src, int sport, in_addr_t dst, int dport)
{
	fprintf(f, "SYN ");
	print_addr(f, src, sport);
	fprintf(f, " -> ");
	print_addr(f, dst, dport);
}

/****************************************************************************
 * The "hash" and "compare" functions for the "lh_new" routine of the lhash
 * library. See lhash(3). We simply return the IP address since it seems to
 * be true that in the VAST majority of cases, there is only ever a single
 * port associated with an IP. We can live with the VERY slight hit of
 * multiple ports (or changing ports) on a few hosts.
 */

static unsigned long calchash(BLOCKAGE *b)
{
	unsigned long rtn;
	rtn = (unsigned long)(b->ip);
	return rtn;
}

static int compare(BLOCKAGE *a, BLOCKAGE *b)
{
	return (
		a->ip == b->ip
		&& (
			a->port == b->port
			|| a->port == -1
			|| b->port == -1
		)
	) ? 0 : 1;
}

/****************************************************************************
 * Add a record to a hash structure; but only if it doesnt already exist.
 * If the record is already present - simply update the "time" stamp
 * contained therein.
 */

int add_to_hash(LHASH *hash, in_addr_t ip, int port)
{
	BLOCKAGE b1;
	BLOCKAGE *b2;

	b1.ip = ip;
	b1.port = port;

	b2 = lh_retrieve(hash, &b1);
	if (b2 != NULL) {
		int rtn = (b2->time == 0);
		b2->time = now;
		b2->probed = FALSE;
		return rtn;
	}

	b2 = malloc(sizeof(BLOCKAGE));
	if (b2 == NULL) {
		error_msg("Out of memory");
		return FALSE;
	}
	b2->ip = b1.ip;
	b2->port = b1.port;
	b2->time = now;
	b2->probed = FALSE;

	lh_insert(hash, b2);
	if (lh_error(hash)) {
		error_msg("lh_insert error");
		return FALSE;
	}
	return TRUE;
}

/****************************************************************************
 * Check to see whether an ip/port pair exists in the specified hash tree.
 * Remember that "port == -1" is a wild card value (see the "compare(..)"
 * function).
 */

int test_hash( LHASH *hash, in_addr_t ip, int port )
{
	BLOCKAGE b;
	int rtn;
	
	b.ip = ip;
	b.port = port;

	rtn = (lh_retrieve(hash, &b) == NULL) ? 0 : 1;
	return rtn;
}

/****************************************************************************
 * Add the specified ip/port pair to the list of addresses that we belieive
 * to refer to FastTrack peers. The "why" string gives a textual
 * description of the reason for the classification - for debug purposes
 * only
 */

int block_this(in_addr_t ip, int port, char *why)
{
	int rtn;

	rtn = add_to_hash(blockages_hash, ip, port);
	if (rtn) {
		if (debug) {
			printf("Block: ");
			print_addr(stdout, ip, port);
			printf("\n");
		}
		if (FTWALL_STORE != NULL) {
			print_addr(FTWALL_STORE,ip,port);
			fprintf(FTWALL_STORE,"\n");
		}
		if (what2log & LOG_IDENTIFIED_PEERS) {
			struct in_addr a;
			a.s_addr = ip;
			write_log("Added %s:%d to peer list (%s)",inet_ntoa(a),port,why);
		}
	}
	return rtn;
}

int block_this_str( uchar *str, char *why )
{
	in_addr_t ip;
	int port;

	ip = inet_addr(strtok((char *)str, ":"));
	port = atoi(strtok(NULL, ":"));
	return block_this(ip, port,why);
}

/****************************************************************************
 * Does the specified ip/port pair refer to something that we have found to
 * be a "peer", and have therefore elected to block?
 */

int is_blocked( in_addr_t ip, int port )
{
	return test_hash(blockages_hash, ip, port);
}

void log_client_added(in_addr_t ip, int port, char *why)
{
	struct in_addr a;
	int id;

	a.s_addr = ip;

	if (what2log & LOG_IDENTIFIED_CLIENTS)
		write_log("Added %s:%d to client list (%s)",inet_ntoa(a),port,why);

	if (active_client_directory != NULL) {
		char *fname = malloc(strlen(active_client_directory) + 30);
		if (fname != NULL) {
			sprintf(fname, "%s/%s", active_client_directory, inet_ntoa(a));
			id = open(fname, O_RDWR | O_CREAT | O_TRUNC, 0666);
			if (id != -1)
				close(id);
			free(fname);
		}
	}
}

void log_client_removed(in_addr_t ip, int port)
{
	struct in_addr a;
	a.s_addr = ip;

	if (what2log & LOG_IDENTIFIED_CLIENTS)
		write_log("Removed %s:%d from client list",inet_ntoa(a),port);

	if (active_client_directory != NULL) {
		char *fname = malloc(strlen(active_client_directory) + 30);
		if (fname != NULL) {
			sprintf(fname, "%s/%s", active_client_directory, inet_ntoa(a));
			unlink(fname);
			free(fname);
		}
	}
}


int add_client(in_addr_t ip, int port, char *why)
{
	int rtn;
	BLOCKAGE b1, *b2;

	b1.ip = ip;
	b1.port = -1;
	b2 = lh_retrieve(clients_hash, &b1);

	/* We assume here that we can only have one fast track port
	 * open on any one client - but that the user may change the
	 * port number in an attempt to "break through". So we track
	 * such changes here. */

	if (b2 != NULL && b2->time != 0) {
		if (b2->port != port && (what2log & LOG_IDENTIFIED_CLIENTS)) {
			struct in_addr a;
			a.s_addr = ip;
			write_log("Client %s changed port from %d to %d (%d)",
					inet_ntoa(a),b2->port,port,why);
		}
		b2->port = port;
		b2->time = now;
		b2->probed = FALSE;
		return TRUE;
	}

	rtn = add_to_hash(clients_hash, ip, port);

	if (rtn)
		log_client_added(ip, port,why);

	return rtn;
}

int add_client_str( uchar *str, char *why )
{
	in_addr_t ip;
	int port;

	ip = inet_addr(strtok((char *)str, ":"));
	port = atoi(strtok(NULL, ":"));
	return add_client(ip, port,why);
}

int is_client(in_addr_t ip, int port)
{
	return test_hash(clients_hash, ip, port);
}

void set_timelock(in_addr_t ip, char *why)
{
	BLOCKAGE b1, *b2;

	if (timelock_secs < 1) return;

	b1.ip = ip;
	b1.port = -1;

	b2 = lh_retrieve(clients_hash, &b1);
	if (b2 != NULL) {
		if (b2->time == 0)
			log_client_added(b2->ip, b2->port,why);
		b2->time = now;
		b2->probed = FALSE;
	}
}

time_t timelock_left(in_addr_t ip)
{
	BLOCKAGE b1;
	BLOCKAGE *b2;
	int rtn;

	if (timelock_secs < 1) return 0;

	b1.ip = ip;
	b1.port = -1;

	b2 = lh_retrieve(clients_hash, &b1);
	if (b2 == NULL)
		return 0;
	return timelock_secs - (now - b2->time);
}

static void send_probe(in_addr_t ip)
{
	struct sockaddr_in addr;
	BLOCKAGE b1, *b2;
	int port;

	b1.ip = ip;
	b1.port = -1;

	b2 = lh_retrieve(clients_hash, &b1);
	if (b2 == NULL)
		return;
	port = b2->port;
	
	addr.sin_family = PF_INET;
	addr.sin_port = htons(port);
	addr.sin_addr.s_addr = ip;

	if (debug) {
		printf("sending fake open to ");
		print_addr(stdout, ip, port);
		printf("\n");
	}

	sendto(probe_socket, probe_pkt, sizeof(probe_pkt), 0, (void *)&addr, sizeof(addr));

	/* The packet we have just sent will be replied to by the client, if it is
	 * still running Fast Track. The reply packet will enter the "INPUT" chain
	 * of the firewall. Provided that the right iptables rule is in place, it
	 * will come to ftwall via the QUEUE target, and will be processed by the
	 * inspect_udp function. This saves listening to the queue AND a socket
	 * at the same time. It's rather a lame way of doing things, and it may
	 * improve in future. */
}

/****************************************************************************
 * Check a client hash record for timelock status. If the timelock has half
 * expired, send a "probe" packet to test for the continued use of the Fast
 * Track software. If the timelock has fully expired, then set the timer to
 * zero (to flag it as inactive) and write a log message. This function is
 * called for every client listed in the hash by the lh_doall function in
 * "check_for_idle_clients".
 */

void check_client(BLOCKAGE *b)
{
	time_t tleft = timelock_secs - (now - b->time);

	/* When the timelock is half expired, send a probe message */

	if (
		tleft < (timelock_secs/2)	/* timelock half or more expired */
		&& !b->probed			/* and .. not probed since timelock was refreshed */
		&& tleft > 0			/* and .. timelock sis till running */
		&& b->time >0			/* and .. host not flagged "inactive" */
	) {
		send_probe(b->ip);
		b->probed = TRUE;
	}

	/* If the time lock expires - note that the client isnt with
	 * us any more */

	if (b->time > 0 && tleft < 0) {
		log_client_removed(b->ip, b->port);
		b->time = 0;
	}
		
}

void scan_for_idle_clients()
{
	static time_t last_scan = 0;
	if (now < last_scan || (now - last_scan) > scan_timer) {
		lh_doall(clients_hash, check_client);
		last_scan = now;
	}
}


static int inspect_tcp( uchar *payload, int len, int hook) {
	int ip_hlen, ip_vers, dport, sport, tcp_hlen;
	struct tcphdr *tcphdr;
	struct iphdr *iphdr;
	uchar *data;
	int datalen;
	time_t tleft;

	if (hook != NF_IP_FORWARD) return TRUE;	/* only interested in FORWARDed packets */

	if (len < 20) return TRUE;		/* not enough for an IP packet */

	iphdr = (struct iphdr *)payload;
	ip_vers = iphdr->version;
	ip_hlen = iphdr->ihl * 4;

	if (ip_vers != 4) return TRUE;		/* not ipV4 */

	tcphdr = (struct tcphdr *)(payload + ip_hlen);
	tcp_hlen = tcphdr->doff * 4;
	dport = ntohs(tcphdr->dest);
	sport = ntohs(tcphdr->source);

	if (dport == 1214 && tcphdr->syn)
		block_this(DST_ADDR, dport,"Outbound TCP/IP SYN to port 1214");

	if (is_blocked( DST_ADDR, dport )) {
		set_timelock(SRC_ADDR, "Is already blocked");
		if (debug) {
			print_syn(stdout, SRC_ADDR, sport, DST_ADDR, dport);
			printf( " blocked\n" );
		}
		return FALSE;
	}

	if (is_client(SRC_ADDR, -1)) {
		tleft = timelock_left(SRC_ADDR);
		if (tleft > 0) {
			if (debug) {
				print_syn(stdout, SRC_ADDR, sport, DST_ADDR, dport);
				printf( " rejected on timelock (%ld)\n", tleft);
			}
			return FALSE;
		}
	}

	/* We're going to look at the data now, so if this isnt a
	 * simple data packet - just accept it without further ado. */

	if (!tcphdr->psh)
		return TRUE;

	data = payload + ip_hlen + tcp_hlen;
	datalen = len - ip_hlen - tcp_hlen;

	if (
		memcmp( data, "GET /", 5) == 0
		|| memcmp( data, "HTTP/1. ", 7) == 0
	) {
		int i;

		uchar netwk_key[] = "X-Kazaa-Network: ";
		int netwk_keylen = sizeof(netwk_key)-1;
		int netwk_key_seen = FALSE;

		uchar snode_key[] = "X-Kazaa-SupernodeIP: ";
		int snode_keylen = sizeof(snode_key)-1;
		uchar *snode_str = NULL;

		uchar client_key[] = "X-Kazaa-IP: ";
		int client_keylen = sizeof(client_key)-1;
		uchar *client_str = NULL;

		for (i=0; i<datalen - netwk_keylen; i++) {
			if ( data[i] == '\n' ) {
				if (memcmp(data+i+1, netwk_key, netwk_keylen) == 0)
					netwk_key_seen = TRUE;
				if (memcmp(data+i+1, snode_key, snode_keylen) == 0)
					snode_str = data+i+1+snode_keylen;
				if (memcmp(data+i+1, client_key, client_keylen) == 0)
					client_str = data+i+1+client_keylen;
			}
		}

		if (netwk_key_seen && data[0] == 'G') {	/* Outgoing GET request */
			if (debug)
				printf("%s",data);
			block_this(DST_ADDR, dport,"Is destination for HTTP GET request");
			set_timelock(SRC_ADDR, "Is source of HTTP GET request");
			if (snode_str != NULL)
				block_this_str(snode_str, "Is Supernode named in HTTP GET request");
			if (client_str != NULL)
				add_client_str(client_str, "Is client named in HTTP GET request");
			return FALSE;
		}

		if (netwk_key_seen && data[0] == 'H') {	/* reply to incoming GET request */
			if (debug)
				printf("%s",data);
			block_this(DST_ADDR, dport,"Is destination for HTTP GET reply");
			set_timelock(SRC_ADDR, "Is source of HTTP GET reply");
			if (snode_str != NULL)
				block_this_str(snode_str,"Is Supernode named in HTTP GET reply");
			if (client_str != NULL)
				block_this_str(client_str,"Is client named in HTTP GET reply");
			return FALSE;
		}
	}

	return TRUE;
}

unsigned short in_cksum(
	unsigned short *ptr1,int nbytes1,
	unsigned short *ptr2,int nbytes2
)
{

	register long sum;	    /* assumes long == 32 bits */
	u_short oddbyte;
	register u_short answer;    /* assumes u_short == 16 bits */

	assert(nbytes1 % 2 == 0);

	sum = 0;

	while (nbytes1 > 1)  {
		sum += *ptr1++;
		nbytes1 -= 2;
	}

	while (nbytes2 > 1)  {
		sum += *ptr2++;
		nbytes2 -= 2;
	}

	if (nbytes2 == 1) {
		oddbyte = 0;	       			 /* make sure top half is zero */
		*((u_char *) &oddbyte) = *(u_char *)ptr2; /* one byte only */
		sum += oddbyte;
	}

	sum  = (sum >> 16) + (sum & 0xffff);	/* add high-16 to low-16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum;				/* ones-complement, then truncate to 16 bits */
	return(answer);
}

static int send_udp(
	in_addr_t srcip, int srcport,
	in_addr_t dstip, int dstport,
	uchar *data, int datalen
)
{
	uchar packet[256];
	static struct sockaddr_in sin;
	static int packet_id = 1234;
	struct iphdr *iphdr = (struct iphdr *)packet;
	struct udphdr *udphdr = (struct udphdr *)(packet + 20);
	struct {
		__u32	saddr;
		__u32	daddr;
		__u8	zero;
		__u8	proto;
		__u16	length;
	} udppseudohdr;
	int rtn;

	assert(datalen+28 <= sizeof(packet));
	assert(sizeof(udppseudohdr) == 12);

	iphdr->version = 4;
	iphdr->ihl = 5;
	iphdr->tos = 0x10;
	iphdr->tot_len = htons(28 + datalen);
	iphdr->id = htons(packet_id ++);
	iphdr->frag_off = 0;
	iphdr->ttl = 0xff;
	iphdr->protocol = IPPROTO_UDP;
	iphdr->check = 0;
	iphdr->saddr = srcip;
	iphdr->daddr = dstip;

	udphdr->source = htons(srcport);
	udphdr->dest = htons(dstport);
	udphdr->len = htons(datalen + 8);
	udphdr->check = 0;

	memcpy(packet+28,data,datalen);

	udppseudohdr.saddr = srcip;
	udppseudohdr.daddr = dstip;
	udppseudohdr.zero = 0;
	udppseudohdr.proto = IPPROTO_UDP;
	udppseudohdr.length = udphdr->len;

	iphdr->check = in_cksum((ushort *)iphdr,20,NULL,0);
	udphdr->check = in_cksum((ushort *)&udppseudohdr,12,(ushort *)(udphdr),datalen+8);

	sin.sin_family = AF_INET;
	sin.sin_port = htons(dstport);
	sin.sin_addr.s_addr = dstip;
	rtn = sendto(raw_socket, packet, datalen+28, 0, (struct sockaddr *)&sin, sizeof(sin));
	return (rtn == datalen+28);
}

static int send_lie( in_addr_t srcip, int sport, in_addr_t dstip, int dport )
{
	send_udp(srcip,sport,dstip,dport,reply_packet,sizeof(reply_packet));
}

static int inspect_udp( uchar *payload, int len, int hook) {
	int ip_hlen, ip_vers, udp_hlen;
	int datalen;
	uchar *data;
	int port;

	if (len < 20) return TRUE;		/* not enough for an IP packet */
	ip_vers = (payload[0] & 0xf0) >> 4;

	if (ip_vers != 4) return TRUE;		/* not ipV4 */
	ip_hlen = (payload[0] & 0x0f) * 4;

	udp_hlen = 8;
	data = payload + ip_hlen + udp_hlen;
	datalen = len - ip_hlen - udp_hlen;

	/*
	 * If the packet is a response to a probe request then refresh the
	 * timelock timer for the client that sent it to us and DROP the
	 * packet. The packet arrives on the INPUT chain.
	 */

	if (
		hook == NF_IP_LOCAL_IN &&
		DST_UDP_PORT == probe_port &&
		is_client(SRC_ADDR,SRC_UDP_PORT)
	) {
		set_timelock(SRC_ADDR, "Responded to probe");
		return FALSE;
	}


	/* Otherwise - we are only interested in checking FORWARDED packets */

	if (hook != NF_IP_FORWARD) return TRUE;


	/* Identify FastTrack's openning-gambit UDP packet */

	if (datalen == 12 && memcmp(data+6, "KaZaA\0", 6) == 0) {
		/* If this is one - flag the peer to which it was sent as blocked */

		block_this( DST_ADDR, DST_UDP_PORT, "Is destination for UDP start packet" );

		/* and take a note that the sender is a home network client */

		add_client( SRC_ADDR, SRC_UDP_PORT, "Is source for UDP start packet" );

		/*
		 * If we are blocking UDP (which we normally should be) then fake
		 * a reply to the client and drop the packet. If we are not blocking
		 * UDP (for Lab testing only) then accept the packet
		 */

		if (block_udp)
			send_lie(DST_ADDR, DST_UDP_PORT, SRC_ADDR, SRC_UDP_PORT);
		set_timelock(SRC_ADDR, "Is source of UDP start packet");
		return block_udp ? FALSE : TRUE;
	}

	/*
	 * If we get here, then the packet is something other than a FastTrack
	 * openning-gambit. In which case we DROP it if it is destined for
	 * something that we know to be a FastTrack peer.
	 */

	if (block_udp && is_blocked( DST_ADDR, DST_UDP_PORT )) {
		set_timelock(SRC_ADDR, "Is source of UDP packet sent to known peer");
		return FALSE;
	}

	/* All other UDP traffic is ACCEPTed */

	return TRUE;
}

void relinfo(FILE *f)
{
	fprintf(f,
		"ftwall "VERSION" by Chris Lowth <chris@lowth.com> "RELDATE"\n"
		"\n"
		"Copyright: (C) 2003, Chris Lowth.\n"
		"License:   GNU General Public License Version 2, June 1991\n"
		"\n"
		"You should have received a copy of the GNU General Public License\n"
		"along with this program; if not, write to the Free Software\n"
		"Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n"
		"\n"
		"This product includes software developed by the OpenSSL Project for use in\n"
		"the OpenSSL Toolkit (http://www.openssl.org/).\n"
	       	"This product includes software written by Eric Young (eay@cryptsoft.com).\n"
	      	"This product includes software written by Tim Hudson (tjh@cryptsoft.com).\n"
		"+--------------------------------------------------------------------+\n"
		"| If you find this software useful, please consider supporting the   |\n"
		"| project in one of the following ways..                             |\n"
		"|                                                                    |\n"
		"|  - Making a donation, however small, via PayPal or purchasing an   |\n"
		"|    item on the authors Amazon.co.uk 'wish list'.                   |\n"
		"|                                                                    |\n"
		"|  - Using Lowth.com as your gateway into Amazon.com or Amazon.co.uk |\n"
		"|    when you shop with them. This costs nothing extra, but Amazon   |\n"
		"|    pays commission to the project on anything you buy in this way. |\n"
		"|                                                                    |\n"
		"| Visit http://www.lowth.com/donate.php for information and links    |\n"
		"| about all these options. Thank you.                                |\n"
		"+--------------------------------------------------------------------+\n"
	);
}

void usage()
{
	relinfo(stderr);
	fprintf(stderr,"=======================================================================\n");
	fprintf(stderr,
		"Usage: ftwall [options]\n"
		"\n"
		" -d            debug and stay in foreground\n"
		" -s            stay in foreground (default: don't)\n"
		" -n            don't block UDP packets (default: do)\n"
		" -b number     set socket buffer size (default: 32k)\n"
		" -f filename   set name of storefile (default: none)\n"
		" -t number     set seconds of timelock. 0 disables (default: 120)\n"
		" -c directory  create/delete files here as clients appear/vanish\n"
		" -l|L what     set events to log to syslog|stderr (default: none)\n"
		"                  u (U) : dropped (accepted) UDP packets\n"
		"                  t (T) : dropped (accepted) TCP/IP packets\n"
		"                  c     : identified green-network clients\n"
		"                  p     : identified public-network peers\n"
		"\n"
	);
	exit(2);
}

void log(ipq_packet_msg_t *m, int dropped)
{
	char msg[1024];
	struct in_addr a;
	char *src_ip, *dst_ip;
	uchar *payload = m->payload;
	struct iphdr *iphdr = (struct iphdr *)payload;
	int ip_hlen = iphdr->ihl * 4;
	char *hookstr = "?";
	char *indev;
	char *outdev;

	if ( (what2log &
		( LOG_DROPPED_UDP | LOG_DROPPED_TCP | LOG_ACCEPTED_UDP | LOG_ACCEPTED_TCP )
	     ) == 0
	)
		return;

	a.s_addr = SRC_ADDR; src_ip = strdup(inet_ntoa(a));
	a.s_addr = DST_ADDR; dst_ip = strdup(inet_ntoa(a));

	switch (m->hook) {
		case NF_IP_PRE_ROUTING:  hookstr = "PRE_ROUTING"; break;
		case NF_IP_LOCAL_IN:     hookstr = "LOCAL_IN"; break;
		case NF_IP_LOCAL_OUT:    hookstr = "LOCAL_OUT"; break;
		case NF_IP_FORWARD:      hookstr = "FORWARD"; break;
		case NF_IP_POST_ROUTING: hookstr = "POST_ROUTING"; break;
	}

	indev = m->indev_name;
	if (indev[0] == '\0') indev = "*";

	outdev = m->outdev_name;
	if (outdev[0] == '\0') outdev = "*";

	if (m->payload[9] == 17 &&
		(
			(dropped && (what2log & LOG_DROPPED_UDP)) ||
			(!dropped && (what2log & LOG_ACCEPTED_UDP))
		)
	) { /* UDP */
		snprintf(msg, sizeof(msg)-1,
			"%s UDP packet (%s) from %s:%s:%d to %s:%s:%d",
			(dropped ? "Drop" : "Accept"),
			hookstr,
			indev, src_ip, SRC_UDP_PORT,
			outdev, dst_ip, DST_UDP_PORT
		);
		write_log( "%s", msg );
	}

	if (m->payload[9] == 6 &&
		(
			(dropped && (what2log & LOG_DROPPED_TCP)) ||
			(!dropped && (what2log & LOG_ACCEPTED_TCP))
		)
	) { /* TCP */
		char flags[256];
		struct tcphdr *tcphdr = (struct tcphdr *)(payload + ip_hlen);
		flags[0] = '\0';
		if (tcphdr->syn) strcat(flags, " SYN" );
		if (tcphdr->rst) strcat(flags, " RST" );
		if (tcphdr->fin) strcat(flags, " FIN" );
		if (tcphdr->ack) strcat(flags, " ACK" );
		if (tcphdr->psh) strcat(flags, " PSH" );
		snprintf(msg, sizeof(msg)-1,
			"%s TCP packet (%s) from %s:%s:%d to %s:%s:%d%s",
			(dropped ? "Drop" : "Accept"),
			hookstr,
			indev, src_ip, SRC_UDP_PORT,
			outdev, dst_ip, DST_UDP_PORT,
			flags
		);
		write_log( "%s", msg );
	}

	free(src_ip);
	free(dst_ip);
}

int main(int argc, char **argv)
{
	int status;
	uchar buf[BUFSIZE];
	struct ipq_handle *h;
	int verdict;
	struct sockaddr_in addr;
	int rtn;
	FILE *store_in;
	int opt;
	FILE *pidfile;
	char *logging_options = "";
	int i;

	now = time(NULL);

	while ((opt = getopt(argc,argv,"dsnf:k:b:t:l:L:c:")) != -1) {
		switch (opt) {
			case 'd': debug = stay_in_foreground = TRUE; break;
			case 's': stay_in_foreground = TRUE; break;
			case 'n': block_udp = FALSE; break;
			case 'f': storefilename = strdup(optarg); break;
			case 'b': sock_bsize = atoi(optarg); break;
			case 't': timelock_secs = atoi(optarg); break;
			case 'l': logging_options = optarg;
				  log2 = LOG_TO_SYSLOG;
				  break;
			case 'L': logging_options = optarg;
				  log2 = LOG_TO_STDERR;
				  stay_in_foreground = TRUE;
				  break;
			case 'c': active_client_directory = optarg; break;
			default: usage();
		}
	}

	if ( optind != argc)
		usage();

	for (i=0; logging_options[i]; i++) {
		switch (logging_options[i]) {
			case 'u': what2log |= LOG_DROPPED_UDP; break;
			case 'U': what2log |= LOG_ACCEPTED_UDP; break;
			case 't': what2log |= LOG_DROPPED_TCP; break;
			case 'T': what2log |= LOG_ACCEPTED_TCP; break;
			case 'p': what2log |= LOG_IDENTIFIED_PEERS; break;
			case 'c': what2log |= LOG_IDENTIFIED_CLIENTS; break;
			default: usage();
		}
	}

	relinfo(stdout);
	printf("\n");

	openlog("ftwall", LOG_PID, LOG_USER);

	blockages_hash = lh_new(calchash, compare);
	clients_hash = lh_new(calchash, compare);

	if (blockages_hash == NULL || clients_hash == NULL)
		die(NULL, "Out of memory");

	if (storefilename != NULL) {
		store_in = fopen( storefilename, "r" );
		if (store_in != NULL) {
			while (fgets((char *)buf, sizeof(buf), store_in) != NULL) {
				block_this_str(buf, "Is listed in store file");
		}
			fclose(store_in);
		}

		FTWALL_STORE = fopen(storefilename, "a" );
		if (FTWALL_STORE == NULL) 
			die(NULL, "Cant open '%s' for writing", storefilename);
		setbuf(FTWALL_STORE, NULL);
	}

	probe_socket = socket(PF_INET, SOCK_DGRAM, 0);
	if (probe_socket == -1)
		die(NULL, "Cant create probe socket");

	addr.sin_family = PF_INET;
	addr.sin_port = htons(probe_port);
	addr.sin_addr.s_addr = inet_addr( "0.0.0.0" );

	rtn = bind(probe_socket, (void *)&addr, sizeof(addr));
	if (rtn != 0) {
		hint = "Is an instance of ftwall already running?";
		die(NULL, "Cant bind probe socket");
	}

	raw_socket = socket(AF_INET, SOCK_RAW, 255);
	if (raw_socket == -1)
		die(NULL, "Cant open RAW socket");

	h = ipq_create_handle(0, PF_INET);
	if (!h) die(h, "ipq_create_handle failed");

	/* Get packet data too, but only upto a small ceiling */
	status = ipq_set_mode(h, IPQ_COPY_PACKET, PAYLOADSIZE);
	if (status < 0) {
		hint = "Is the ip_queue kernel module loaded? Try 'modprobe ip_queue'.";
		die(h, "ipq_set_mode failed");
	}

	/* Set the socket buffer sizes */
	if (
		setsockopt(h->fd, SOL_SOCKET, SO_RCVBUF, &sock_bsize, sizeof(sock_bsize)) != 0 || 
		setsockopt(h->fd, SOL_SOCKET, SO_SNDBUF, &sock_bsize, sizeof(sock_bsize)) != 0
	)
		die(h, "setsockopt failed" );

	if ( active_client_directory != NULL )
		mkdir(active_client_directory, 0755);

	if ( !stay_in_foreground ) {
		if (fork() == 0) {
			signal(SIGHUP, SIG_IGN);

			stdio_is_null = TRUE;
			close(0); open("/dev/null", O_RDWR);
			close(1); open("/dev/null", O_RDWR);
			close(2); open("/dev/null", O_RDWR);

			pidfile = fopen("/var/run/ftwall.pid", "w");
			if (pidfile != NULL) {
				fprintf(pidfile, "%d\n", getpid());
				fclose(pidfile);
			}
		} else {
			return 0;
		}
	}

	for ( ; ; ) {
		memset(buf, 0, BUFSIZE);
		status = ipq_read(h, buf, BUFSIZE, scan_timer * 1024 * 1024);

		now = time(NULL);

		if (status == 0) {	/* Time out */
			scan_for_idle_clients();
			continue;
		}

		if (status < 0) {
     			ipq_perror("irq_read");
			continue;
		}

		scan_for_idle_clients();

		switch (ipq_message_type(buf)) {

		case NLMSG_ERROR: {
			error_msg("Received error message %d", ipq_get_msgerr(buf));
			break;
		}

		case IPQM_PACKET: {
			ipq_packet_msg_t *m = ipq_get_packet(buf);
			if (m->payload[9] == 6)
				verdict = inspect_tcp(m->payload, m->data_len, m->hook);
			else if (m->payload[9] == 17)
				verdict = inspect_udp(m->payload, m->data_len, m->hook);
			else
				verdict = TRUE;

			log(m, verdict == NF_DROP);

			status = ipq_set_verdict(h, m->packet_id,
				verdict ? NF_ACCEPT : NF_DROP, 0, NULL
			);

			if (status < 0){
				ipq_perror("irq_set_verdict");
				continue;
			}
			break;
		}

		default:
			error_msg("Unknown message type!");
			break;
		}
	}

	ipq_destroy_handle(h);

	return 0;
}
#endif
